#! /usr/bin/python3

import os, sys, math

if sys.version_info[0] > 3:
    raise Exception("Python 3 is required.")

import configparser
import linuxcnc
from PyQt5 import QtCore, uic
from PyQt5.QtWidgets import (QApplication, QMainWindow, QLabel, QCheckBox, QRadioButton,
QMessageBox, QFileDialog, QMenu)
from PyQt5.QtCore import Qt, QTimer, QEvent

#print(os.environ)
IN_AXIS = os.environ["LINUXCNCVERSION"]
FILE_PATH = os.environ["LINUXCNC_NCFILES_DIR"]
print('FilePath is ', FILE_PATH)

if  os.path.split(os.environ.get("PYTHONPATH")) == '/usr/bin':
    GUI_PATH = '/usr/lib/python3/dist-packages'
    print('Installed')

else: 
    pythonPath = os.environ.get("PYTHONPATH")
    print ('PythonPath is', pythonPath)
    GUI_PATH = pythonPath


class MainWindow(QMainWindow):
    def __init__(self):
        super(MainWindow, self).__init__()
        self.s = linuxcnc.stat() # create a connection to the status channel
        try: # make sure linuxcnc is running
            self.s.poll()
        except linuxcnc.error:
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Critical)
            msg.setWindowTitle('Error')
            msg.setText('LinuxCNC is not running')
            msg.setInformativeText('Start LinuxCNC first.')
            msg.setStandardButtons(QMessageBox.Ok)
            msg.exec_()
            exit()
        self.filePath = FILE_PATH
        uic.loadUi(os.path.join('libemclog/emclog.ui'), self)
        self.qclip = QApplication.clipboard()
        self.gcodeLW.installEventFilter(self)
        self.axes = [(i) for i in range(9)if self.s.axis_mask & (1<<i)]

        self.axes_letters = ['X', 'Y', 'Z', 'A', 'B', 'C', 'U', 'V', 'W']
        # initialize variables for the axis-checkboxes and position-labels
        for i in range(9):
            setattr(self, 'axisCB_' + str(i), None)                    
            setattr(self, 'positionLB_' + str(i), None)
        # organize the checkboxes and labels in two lists
        self.axis_cbs = [self.axisCB_0, self.axisCB_1, self.axisCB_2, self.axisCB_3,
                         self.axisCB_4, self.axisCB_5, self.axisCB_6, self.axisCB_7, self.axisCB_8]
        self.position_lbs = [self.positionLB_0, self.positionLB_1, self.positionLB_2, self.positionLB_3,
                             self.positionLB_4, self.positionLB_5, self.positionLB_6, self.positionLB_7,
                             self.positionLB_8]
        # for each active axis we create a QCheckBox and a QLabel widget using the respective list 
        # and arrange them in a grid
        (row, col) = (0,0)
        self.activeAxesdLayout.setSpacing(2)          
        for i in range(9):
            if self.s.axis_mask & (1<<i):
                self.axis_cbs[i] = QCheckBox(self.axes_letters[i]+':')
                self.activeAxesdLayout.addWidget(self.axis_cbs[i],row,col)
                self.axis_cbs[i].setChecked(True)
                self.position_lbs[i] = QLabel('pos'+str(i))
                self.activeAxesdLayout.addWidget(self.position_lbs[i],row,col+1)
                self.position_lbs[i].setMinimumWidth(60)
                row += 1
                if row == 6: 
                    (col,row) = (2,0)

        self.setupConnections()
        self.timer = QTimer()
        self.timer.timeout.connect(self.update)
        self.timer.start(100)
        self.recordTimer = QTimer()
        self.recordTimer.timeout.connect(self.log)
        self.lastPosition = []
        self.dinLog = False
        self.dinInput = self.dioSB.value()
        self.pref_path = os.path.join(os.path.expanduser("~"), '.emclog_preferences')
        self.config = configparser.ConfigParser()
        self.logToLine = None
        self.setStyleSheet('QListWidget::item:selected{background: rgb(0,127,127);}')
        self.setupGUI()
        self.show()

    def eventFilter(self, source, event):
        if (event.type() == QEvent.ContextMenu and source is self.gcodeLW):
            contextMenu = QMenu()
            addLineAbove = contextMenu.addAction('Add Line Above')
            addLineBelow = contextMenu.addAction('Add Line Below')
            logToLine = contextMenu.addAction('Log to this Line')
            deleteLine = contextMenu.addAction('Delete Line')
            action = contextMenu.exec_(self.mapToGlobal(event.pos()))
            if action == addLineAbove:
                self.gcodeLW.insertItem(self.gcodeLW.currentRow(),';new line')
            if action == addLineBelow:
                self.gcodeLW.insertItem(self.gcodeLW.currentRow()+1,';new line')
            if action == logToLine:
                self.logToLine = self.gcodeLW.currentRow()
                self.setStyleSheet('QListWidget::item:selected{background: rgb(254,127,0);}')
            if action == deleteLine:
                self.gcodeLW.takeItem(self.gcodeLW.currentRow())
            return True
        return super(MainWindow, self).eventFilter(source, event)

    def setupConnections(self):
        self.actionOpen.triggered.connect(self.OpenFile)
        self.actionSave.triggered.connect(self.SaveFile)
        self.actionSave_As.triggered.connect(self.SaveFileAs)
        self.actionSavePreferences.triggered.connect(self.SavePreferences)

        self.actionExit.triggered.connect(self.exit)
        self.logPB.clicked.connect(self.log)
        self.actionCopy.triggered.connect(self.copy)
        self.startPB.clicked.connect(self.record)
        self.stopPB.clicked.connect(self.record)
        self.dioSB.valueChanged.connect(self.changeInput)
        self.gcodeLW.itemDoubleClicked.connect(self.DoubleClicked)

    def DoubleClicked(self,item):
        self.logToLine = None
        #QMessageBox.information(self, "ListWidget", "You clicked: "+item.text())
        item.setFlags(item.flags() | QtCore.Qt.ItemIsEditable)
        self.gcodeLW.editItem(item)
        self.setStyleSheet('QListWidget::item:selected{background: rgb(0,127,127);}')

    def changeInput(self):
        self.dinInput = self.dioSB.value()

    def OpenFile(self):
        fileName = QFileDialog.getOpenFileName(self,
        caption="Select a G code File",
        directory = self.filePath,
        filter='*.ngc',
        options=QFileDialog.DontUseNativeDialog,)
        if fileName[0]:
            self.gcodeLW.clear()
            with open(fileName[0], 'r') as f:
                for line in f:
                    self.gcodeLW.addItem(line.strip('\n'))
            # update the file path 
            self.filePath = fileName[0]
            self.gcodeLB.setText(fileName[0])

    def SaveFile(self):
        if os.path.isfile(self.filePath):
            print('Saving to ', self.filePath)
            gcode = '\n'.join(self.gcodeLW.item(i).text() for i in range(self.gcodeLW.count()))
            with open(self.filePath, 'w') as f:
                f.write(gcode)
        else:
            self.SaveFileAs()

    def SaveFileAs(self):
        if self.filePath is tuple:
            path = self.filePath[0]
        else:
            path = self.filePath
        fileName, _ = QFileDialog.getSaveFileName(self,
        caption="Save G Code",
        directory = path,
        filter=("Gcode (*.ngc)"),
        options=QFileDialog.DontUseNativeDialog)
        if fileName:
            gcode = '\n'.join(self.gcodeLW.item(i).text() for i in range(self.gcodeLW.count()))
            with open(fileName, 'w') as f:
                f.write(gcode)
            # update the file path 
            self.filePath = fileName
            self.gcodeLB.setText(fileName)
 
    def SavePreferences(self):
        if not os.path.exists(self.pref_path):
            self.config.add_section('MAIN')
        self.config.set('MAIN','dio', str(self.dioSB.value()))
        with open(self.pref_path, 'w') as cfgfile:
            self.config.write(cfgfile)
        #    f.write(str(self.dioSB.value()))

    def setupGUI(self):
        self.positionCB.addItem('Relative', 'relative')
        self.positionCB.addItem('Absolute', 'absolute')

        # check axes that are and disable axes that are not
        for i in range(9):
            if getattr(self, 'axisCB_' + str(i)) != None:
                print('axisCB_' + str(i))
                getattr(self, 'axisCB_' + str(i)).setChecked(i in self.axes)

        if os.path.exists(self.pref_path):
            self.config.read(self.pref_path)
            self.dioSB.setValue(self.config.getint('MAIN', 'dio'))

    def record(self):
        if self.startPB.isChecked():
            print('Starting {}'.format(self.intervalSB.value()))
            timerInterval = self.intervalSB.value() * 1000
            self.recordTimer.start(timerInterval)

        elif self.stopPB.isChecked():
            print('Stopping')
            self.recordTimer.stop()

    def log(self):
        axes = []
        #for checkbox in self.axesGB.findChildren(QCheckBox): # get axes list
        for cb in range(len(self.axis_cbs)): # get axes list
            if self.axis_cbs[cb] != None and self.axis_cbs[cb].isChecked():
                axes.append(cb)
        gcode = []
        currentPosition = []
        for radio in self.moveGB.findChildren(QRadioButton):
            if radio.isChecked(): # add the move type
                gcode.append(str(radio.property('gcode')) + ' ')
                moveType = str(radio.property('gcode'))
        for axis in axes: # add each axis position
            position = self.position_lbs[axis].text()
            currentPosition.append(float(position))
            gcode.append(self.axes_letters[axis] + position + ' ')

        if moveType in ['G2', 'G3']:
            if self.arcRadiusLE.text() == '':
                self.mbox('{} moves require an arc radius'.format(moveType))
                return
            if len(self.lastPosition) == 0:
                self.mbox('A G0 or G1 move must be done before a {} move'
                .format(moveType))
            dx = currentPosition[0] - self.lastPosition[0]
            dy = currentPosition[1] - self.lastPosition[1]
            if dx == 0 and dy == 0:
                self.mbox('{} move needs a different end point'.format(moveType))
                return
            # calculate angle of travel
            angle = math.atan2(dy,dx)
            # calculate the midpoint of the segment
            (dxMid,dyMid) = (dx/2, dy/2)
            # calculate the distance traveled 
            distance = math.sqrt(dx**2 + dy**2)
            # get requested radius
            radius = float(self.arcRadiusLE.text())
            if radius < (distance / 2):
                self.mbox('Radius can not be smaller than {0:0.4f}'.format(distance/2))
                return
            # calculate the distance of the arc-center from the mid-point
            center_offset = math.sqrt(radius**2-(distance/2)**2)

        if moveType == 'G2':
            # arc-center is offset -90° from midpoint of the traveled segment 
            offset_angle = angle-math.pi/2
            i = dxMid + center_offset * math.cos(offset_angle)
            j = dyMid + center_offset * math.sin(offset_angle)
            gcode.append(' I{0:.{2}f} J{1:.{2}f}'.format(i, j, self.precisionSB.value()))
        elif moveType == 'G3':
            # arc-center is offset +90° from midpoint of the traveled segment
            offset_angle = angle+math.pi/2
            i = dxMid + center_offset * math.cos(offset_angle)
            j = dyMid + center_offset * math.sin(offset_angle)
            gcode.append(' I{0:.{2}f} J{1:.{2}f}'.format(i, j, self.precisionSB.value()))

        if moveType in ['G1', 'G2', 'G3']: # check for a feed rate
            feedMatch = self.gcodeLW.findItems('F', Qt.MatchContains)
            if len(feedMatch) > 0: # check last feed rate to see if it is different
                lastMatch =  str(feedMatch[-1].text()).split()
                if lastMatch[-1][1:] != self.feedLE.text():
                    gcode.append(' F{}'.format(str(self.feedLE.text())))
            if not self.gcodeLW.findItems('F', Qt.MatchContains):
                if self.feedLE.text():
                    gcode.append(' F{}'.format(str(self.feedLE.text())))
                else:
                    self.mbox('A feed rate must be entered for a {} move'.format(moveType))
                    return
        if self.logToLine != None:
            # insert the logged values 
            self.gcodeLW.insertItem(self.logToLine, ''.join(gcode))
            # remove the original line that is now below
            self.gcodeLW.takeItem(self.logToLine+1)
            self.setStyleSheet('QListWidget::item:selected{background: rgb(0,127,127);}')
        else:
            self.gcodeLW.addItem(''.join(gcode))
        self.lastPosition = []
        for axis in axes:
            self.lastPosition.append(float(self.position_lbs[axis].text()))
        self.logToLine = None

    def update(self):
        self.s.poll()
        if self.positionCB.currentData() == 'relative':
            # sum the offsets with a negative sign
            offsets = tuple(-sum(i) for i in zip(self.s.g5x_offset,self.s.g92_offset))
            display = tuple(sum(i) for i in zip(offsets,self.s.actual_position))
        else:
            display = self.s.actual_position
        for i in self.axes:
            if self.position_lbs[i] != None:
                self.position_lbs[i].setText('{0:0.{1}f}'.format(display[i],
                    self.precisionSB.value()))
        if self.s.din[self.dinInput] and self.dioCB.isChecked() and not self.dinLog:
            self.log()
            self.dinLog = True
            print('remote log')
        elif not self.s.din[self.dinInput] and self.dinLog:
            self.dinLog = False

    def copy(self):
        items = []
        gcode = [str(self.gcodeLW.item(i).text()) for i in range(self.gcodeLW.count())]
        self.qclip.setText('\n'.join(gcode))

    def mbox(self, message):
        msg = QMessageBox()
        msg.setIcon(QMessageBox.Critical)
        msg.setWindowTitle('Error')
        msg.setText(message)
        msg.setStandardButtons(QMessageBox.Ok)
        msg.exec_()

    def exit(self):
        exit()

def main():
    app = QApplication(sys.argv)
    ex = MainWindow()
    sys.exit(app.exec_())

if __name__ == "__main__":
    main()
